import pytest

import numpy as np

from pysprint.core.methods import SPPMethod
from pysprint.core.bases import Dataset


@pytest.fixture()
def construct_ifg_sequence():
    x = np.arange(1000)
    y = np.sin(x)

    # setting multiple SPP
    d1 = Dataset(x, y)
    d1.set_SPP_data(10, [40, 45])

    # simple
    d2 = Dataset(x, y)
    d2.set_SPP_data(12, 13)

    # setting 3 SPP
    d3 = Dataset(x, y)
    d3.set_SPP_data(14, [14, 500, 43])

    # setting SPP directly
    d4 = Dataset(x, y)
    d4.delay = 40
    d4.positions = np.array([45.8, 54])

    # handle duplicates
    d5 = Dataset(x, y)
    d5.delay = 40
    d5.positions = [5.8, 40]

    # ambiguous positions
    d6 = Dataset(x, y)
    d6.delay = 40
    d6.positions = 45.8, 54

    return d1, d2, d3, d4, d5, d6


def test_collection(construct_ifg_sequence):
    """
    Here we test the basic functionality of calculate_from_ifg.
    """
    d1, d2, d3, d4, _, _ = construct_ifg_sequence
    disp, _, _ = SPPMethod.calculate_from_ifg(
        [d1, d2, d3, d4], reference_point=2, order=5
    )
    np.testing.assert_almost_equal(15.272952, disp[0], decimal=5)
    np.testing.assert_almost_equal(-12.27234, disp[1], decimal=5)
    np.testing.assert_almost_equal(0.76698, disp[2], decimal=5)
    np.testing.assert_almost_equal(-0.02171, disp[3], decimal=5)
    np.testing.assert_almost_equal(0.00014, disp[4], decimal=5)


def test_spp_typeerror():
    with pytest.raises(TypeError):
        SPPMethod.calculate_from_ifg([10, 11], reference_point=2, order=2)


def test_duplicate_entries(construct_ifg_sequence):
    """
    Here we test that duplicated delay values (apart from multiple SPP positions)
    are correctly identified and ValueError is raised.
    """
    d1, d2, d3, d4, d5, d6 = construct_ifg_sequence
    with pytest.raises(ValueError):
        SPPMethod.calculate_from_ifg([d1, d2, d3, d4, d5], reference_point=2, order=4)


def test_ambiguous_positions(construct_ifg_sequence):
    """
    Here we test if ambiguous positions (defining them as tuple) are handled correctly.
    """
    d1, d2, d3, d4, d5, d6 = construct_ifg_sequence
    disp, _, _ = SPPMethod.calculate_from_ifg(
        [d1, d2, d3, d6], reference_point=2, order=5
    )
    np.testing.assert_almost_equal(15.272952, disp[0], decimal=5)
    np.testing.assert_almost_equal(-12.27234, disp[1], decimal=5)
    np.testing.assert_almost_equal(0.76698, disp[2], decimal=5)
    np.testing.assert_almost_equal(-0.02171, disp[3], decimal=5)
    np.testing.assert_almost_equal(0.00014, disp[4], decimal=5)


# TODO : Make more verbose tests for that
def test_spp_append():
    s = SPPMethod(["test_rawparsing.trt"])
    s.append("test_peak.txt")
    assert len(s) == 2



__all__ = [
    "PySprintWarning",
    "DatasetError",
    "InterpolationWarning",
    "FourierWarning",
    "NotCalculatedException",
]


class PySprintWarning(Warning):
    """
    Base pysprint warning class.
    """

    pass


class InterpolationWarning(PySprintWarning):
    """
    This warning is raised when a function applies linear
    interpolation on the data.
    """

    pass


class FourierWarning(PySprintWarning):
    """
    This warning is raised when FFT is called first instead of IFFT.
    Later on it will be improved.
    For more details see help(pysprint.FFTMethod.calculate)
    """

    pass


class DatasetError(Exception):
    """
    This error is raised when invalid type of data encountered
    when initializing a dataset or inherited object.
    """
    pass


class NotCalculatedException(ValueError):
    """
    This error is raised when a function is not available yet
    because something needs to be calculated before.
    """
    pass



from math import factorial

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit

from pysprint.config import _get_config_value
from pysprint.utils import find_nearest, run_from_ipython, _progress
from pysprint.core._functions import _cosfit_config, cos_fit1


class FitOptimizer:
    """Class to help achieve better fitting results."""

    def __init__(self, x, y, ref, sam, reference_point, max_order=3):
        self.x = x
        self.y = y
        self.ref = ref
        self.sam = sam
        if not isinstance(self.x, np.ndarray):
            self.x = np.asarray(self.x, dtype=np.float64)
        if not isinstance(self.y, np.ndarray):
            self.y = np.asarray(self.y, dtype=np.float64)
        if not isinstance(self.ref, np.ndarray):
            try:
                self.ref = np.asarray(self.ref, dtype=np.float64)
            except (ValueError, TypeError):
                pass
                # we ignore because it might be optional
                # if y is already normalized
        if not isinstance(self.sam, np.ndarray):
            try:
                self.sam = np.asarray(self.sam, dtype=np.float64)
            except (ValueError, TypeError):
                pass
                # we ignore because it might be optional
                # if y is already normalized
        if len(self.ref) == 0:
            self._y_norm = self.y
        else:
            self._y_norm = (self.y - self.ref - self.sam) / (
                2 * np.sqrt(self.sam * self.ref)
            )
        self.reference_point = reference_point
        self.x -= self.reference_point
        self.func = cos_fit1
        self.p0 = [1, 1, 1, 1]
        self.popt = self.p0
        self._init_set = False
        self.counter = 0
        self.curr_order = 1
        self.max_order = max_order
        self.rest = None
        self.figure = plt.figure()

    def set_final_guess(self, GD, GDD=None, TOD=None, FOD=None, QOD=None, SOD=None):
        self.p0[3] = GD
        self.rest = []
        if GDD is not None:
            self.rest.append(GDD / 2)
        if TOD is not None:
            self.rest.append(TOD / 6)
        if FOD is not None:
            self.rest.append(FOD / 24)
        if QOD is not None:
            self.rest.append(QOD / 120)
        if SOD is not None:
            self.rest.append(SOD / 720)

    @property
    def user_guess(self):
        return self.rest

    # TODO: Make this update the plot, not redraw
    def update_plot(self):
        plt.clf()
        plt.plot(self.x, self._y_norm)
        plt.plot(self._x_curr, self._y_curr, "k", label="Affected data")
        plt.plot(
            self._x_curr, self.func(self._x_curr, *self.popt), "r--", label="Fit",
        )
        plt.grid()
        plt.draw()
        plt.xlabel(r"$\Delta\omega\, [PHz]$")
        plt.ylabel("I")
        plt.show()

    def set_initial_region(self, percent):
        """ Determines the initial region to fit"""
        self._init_set = True
        if percent < 0 or percent > 100:
            raise ValueError("percent must satisfy percent > 0 or percent < 100.")
        _, idx = find_nearest(self.x, 0)
        self._upper_bound = np.floor(idx + (percent / 2) * (len(self.x) + 1))
        self._lower_bound = np.floor(idx - (percent / 2) * (len(self.x) + 1))
        self._upper_bound = self._upper_bound.astype(int)
        self._lower_bound = self._lower_bound.astype(int)
        if self._lower_bound < 0:
            self._lower_bound = 0
        if self._upper_bound > len(self.x):
            self._upper_bound = len(self.x)
        self._x_curr = self.x[self._lower_bound:self._upper_bound]
        self._y_curr = self._y_norm[self._lower_bound:self._upper_bound]

    def _step_up_func(self):
        """
        Change the function to fit. Starts with first order
        and stepping up until max_order.
        """
        if self.curr_order == self.max_order:
            return
        try:
            self.func = _cosfit_config[self.curr_order + 1]
        except KeyError as e:
            raise ValueError("Order must be in [1, 6].") from e
        try:
            self.p0 = np.append(self.p0, self.rest[self.curr_order - 1])
        except (IndexError, ValueError):
            self.p0 = np.append(self.p0, 1)
        self.curr_order += 1

    def _extend_region(self, extend_by=0.1):
        """ Extends region of fit"""
        self._new_lower = np.floor(self._lower_bound - extend_by * len(self.x))
        self._new_upper = np.floor(self._upper_bound + extend_by * len(self.x))
        self._new_lower = self._new_lower.astype(int)
        self._new_upper = self._new_upper.astype(int)
        self._lower_bound = self._new_lower
        self._upper_bound = self._new_upper
        if self._new_lower < 0:
            self._new_lower = 0
        if self._new_upper > len(self.x):
            self._new_upper = len(self.x)
        self._x_curr = self.x[self._new_lower:self._new_upper]
        self._y_curr = self._y_norm[self._new_lower:self._new_upper]

    def _finetune(self):
        """
        Changes the last parameter randomly, we might get lucky..
        """
        self.p0[-1] = float(np.random.uniform(-1, 1, 1)) * self.p0[-1]

    def _fit(self):
        try:
            if len(self._x_curr) == len(self.x):
                return True
            self.popt, self.pcov = curve_fit(
                self.func, self._x_curr, self._y_curr, p0=self.p0
            )
            self.p0 = self.popt
        except RuntimeError:
            self._finetune()
            self.popt, self.pcov = curve_fit(
                self.func, self._x_curr, self._y_curr, p0=self.p0
            )

    def _fit_goodness(self):
        """
        Coeffitient of determination a.k.a. r^2
        """
        residuals = self._y_curr - self.func(self._x_curr, *self.popt)
        ss_res = np.sum(residuals ** 2)
        ss_tot = np.sum((self._y_curr - np.mean(self._y_curr)) ** 2)
        return 1 - (ss_res / ss_tot)

    # TODO : this should be pprint_disp
    def result_wrapper(self):
        precision = _get_config_value("precision")
        labels = ("GD", "GDD", "TOD", "FOD", "QOD", "SOD")
        params = self.p0[3:]
        for i, (label, param) in enumerate(zip(labels, params)):
            if run_from_ipython():
                from IPython.display import display, Math # noqa

                display(
                    Math(
                        f"{label} = {(params[i] * factorial(i + 1)):.{precision}f} \\ fs^{i + 1}"
                    )
                )
            else:
                print(f"{label} = {(params[i] * factorial(i + 1)):.{precision}f} fs^{i + 1}")

    @_progress
    def run(self, r_extend_by, r_threshold, max_tries=5000, show_endpoint=True):
        precision = _get_config_value("precision")
        if not self._init_set:
            raise ValueError("Set the initial conditions.")
        self._fit()
        while self._fit_goodness() > r_threshold:

            # self.figure.savefig(f'{self.counter}.eps')
            # self.update_plot()

            self._extend_region(r_extend_by)
            self._fit()
            self.counter += 1
            self._step_up_func()
            if self._fit() is True:
                if show_endpoint:
                    self.update_plot()
                    # self.figure.savefig(f'{self.counter}.eps')

                self.result_wrapper()
                if run_from_ipython():
                    from IPython.display import display, Math # noqa

                    display(Math(f"with \\ R^2 = {(self._fit_goodness()):.{precision}f}."))
                else:
                    print(f"with R^2 = {(self._fit_goodness()):.{precision}f}.")
                return self.popt
            if self.counter == max_tries:
                if show_endpoint:
                    self.update_plot()
                print(
                    f"""Max tries (currently set to {max_tries}) reached without finding a suitable fit."""
                )
                return np.zeros_like(self.popt)

        while self._fit_goodness() < r_threshold:
            self._fit()
            # self._finetune()
            self.counter += 1
            if self.counter == max_tries:
                if show_endpoint:
                    self.update_plot()
                print(
                    f"""Max tries (currently set to {max_tries}) reached without finding a suitable fit."""
                )
                return np.zeros_like(self.popt)



"""
Smoke testing utility to check that the notebooks aren't broken.
"""
import ast
import atexit
from contextlib import contextmanager
import glob
import os
import pathlib
import subprocess
import sys
import traceback

HEADER = "import matplotlib\nmatplotlib.use('Agg')\n"

SKIP_NAMES = ("doc_requirements.txt",)

IN_CI = "CI" in os.environ.keys() or 'TF_BUILD' in os.environ.keys()


class MplBackendRewriter(ast.NodeTransformer):
    """
    Simply rewrite ps.interactive context managers to use non-GUI backend.
    """

    def visit_With(self, node):
        self.generic_visit(node)
        for leaf in node.items:
            if leaf.context_expr.func.attr == "interactive":
                leaf.context_expr.args = [ast.Str(s='Agg')]
        ast.fix_missing_locations(node)
        return node


class ExitHooks:
    """
    Patch for sys.excepthook because atexit shadows the exit status.
    """

    def __init__(self):
        self.exit_code = None
        self.exception = None

    def hook(self):
        self._orig_exit = sys.exit
        sys.exit = self.exit
        sys.excepthook = self.exc_handler

    def exit(self, code=0):
        self.exit_code = code
        self._orig_exit(code)

    def exc_handler(self, exc_type, exc, *args):
        self.exception = exc


@contextmanager
def redirected_output(new_stdout=None, new_stderr=None):
    save_stdout = sys.stdout
    save_stderr = sys.stderr
    if new_stdout is not None:
        sys.stdout = new_stdout
    if new_stderr is not None:
        sys.stderr = new_stderr
    try:
        yield
    finally:
        sys.stdout = save_stdout
        sys.stderr = save_stderr


def exec_notebooks(test_dir, log_path='notebooktest.log'):
    # Convert notebooks to .py files
    # This shouldn't be subprocess.call, we should use nbconvert API.
    if not IN_CI:
        subprocess.call(f"jupyter nbconvert --to script {test_dir}\\hu_*.ipynb")

    # find the converted files
    test_files = glob.glob(os.path.join(test_dir, 'hu_*.py'))
    test_files.sort()

    passed = []
    failed = []

    with open(log_path, 'w') as f:

        # Capture the output and traceback and write it to the log file
        with redirected_output(new_stdout=f, new_stderr=f):

            for fname in test_files:

                print(f">>> Rewriting interactive calls in '{fname}'.")

                try:
                    with open(fname, 'r', encoding="utf-8") as fp:
                        data = fp.readlines()

                    # Use the "Agg" backend outside context managers in
                    # the converted notebook files.
                    data = HEADER + ''.join(data)
                    tree = ast.parse(data)

                    # Rewrite all the interactive context managers to
                    # ensure matplotlib is not blocking execution.
                    MplBackendRewriter().visit(tree)
                    ast.fix_missing_locations(tree)

                    print(f">>> Rewrite done.")
                    print(f">>> Executing '{fname}.'")

                    exec(
                        compile(tree, filename="out", mode="exec"),
                        {"__name__": "__main__"}
                    )

                    print(f">>> Passed {fname}.")
                    passed.append(fname)

                except Exception:

                    traceback.print_exc()
                    failed.append(fname)

    print(f">>> Passed {len(passed)}/2")
    print(f">>> Expected to fail {len(failed)}/6")
    print(f">>> Log created at {log_path}.")

    if len(passed) != 2 or len(failed) != 6:
        return 1
    print(f">>> Notebook tests passed.")
    return 0


def cleanup(test_path):
    if hooks.exit_code is not None:
        if hooks.exit_code == 0:
            pyfiles = pathlib.Path(*test_path).absolute().glob("hu_*.py")
            txtfiles = pathlib.Path(*test_path).absolute().glob("*.txt")

            print("Removing .py files:")
            for file in pyfiles:
                file.unlink()
                print(f"Removed {file}.")

            print("Removing .txt files:")
            for file in txtfiles:
                if file.name not in SKIP_NAMES:
                    file.unlink()
                    print(f"Removed {file}.")
            return 0
        else:
            print(
                f"exec_notebooks exited with non-zero ({hooks.exit_code}) exit status. Failed."
            )
            return hooks.exit_code
    elif hooks.exception is not None:
        print(f"Process terminated by exception: {hooks.exception}")
        return 1
    else:
        return 1


hooks = ExitHooks()
hooks.hook()

if __name__ == '__main__':

    # Do a cleanup if outside CI services
    if not IN_CI:
        atexit.register(cleanup, test_path=sys.argv[1:])

    sys.exit(exec_notebooks(*sys.argv[1:]))



# Configuration file for the Sphinx documentation builder.

# -- Path setup --------------------------------------------------------------

autodoc_mock_imports = [
    'pandas',
    'pytest',
    'numpy',
    'scipy',
    'lmfit',
    'matplotlib',
    'numba',
    'matplotlib.rcParams'
]

import os
import sys
sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('../pysprint'))
sys.path.insert(0, os.path.abspath('../pysprint/core/bases'))
sys.path.insert(0, os.path.abspath('../pysprint/core/methods'))
sys.path.insert(0, os.path.abspath('../pysprint/utils'))

# -- Project information -----------------------------------------------------

project = 'Pysprint'
copyright = '2020, Peter Leeh'
author = 'Peter Leeh'

release = '0.12.4'

autodoc_default_options = {
    'special-members': '__init__'
}

# -- General configuration ---------------------------------------------------
master_doc = 'index'

extensions = [
    'recommonmark',
    'nbsphinx',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.mathjax',
]

templates_path = ['_templates']


exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store',
    'pysprint.tests.rst',
    'pysprint.mpl_tools.rst',
    'pysprint.bases.io.rst',
    '**.ipynb_checkpoints'
]


intersphinx_mapping = {
    'numpy': ('https://numpy.org/doc/stable/', None),
    'matplotlib': ('https://matplotlib.org/', None),
    'pandas': ('https://pandas.pydata.org/', None),
    'scipy': ('http://docs.scipy.org/doc/scipy/reference/', None)
}

# -- Options for HTML output -------------------------------------------------

html_theme = 'sphinx_rtd_theme'

html_static_path = ['_static']

autodoc_member_order = 'bysource'
add_module_names = False


def maybe_skip_member(app, what, name, obj, skip, options):
    EXCLUDE_CLASSES = (
        'DatasetBase', 'DatasetApply', 'pysprint.core.bases.apply',
        'pysprint.core.bases.dataset_base', '**/*dataset_base*'
    )
    exclude = name in EXCLUDE_CLASSES
    return skip or exclude


def setup(app):
    app.connect('autodoc-skip-member', maybe_skip_member)



from unittest.mock import patch

import pytest
import numpy as np

from pysprint import CosFitMethod


@pytest.fixture()
def make_obj():
    return CosFitMethod(np.arange(100), np.arange(100))


def test_offs(make_obj):
    c = make_obj
    c.adjust_offset(10)
    assert c.params[0] == 10


def test_amp(make_obj):
    c = make_obj
    c.adjust_amplitude(10)
    assert c.params[1] == 10


def test_disp_guesses(make_obj):
    c = make_obj
    c.guess_GD(10)
    c.guess_GDD(20)
    c.guess_TOD(30)
    c.guess_FOD(40)
    c.guess_QOD(50)
    c.guess_SOD(60)
    assert c.params[3] == 10
    assert c.params[4] == 20 / 2
    assert c.params[5] == 30 / 6
    assert c.params[6] == 40 / 24
    assert c.params[7] == 50 / 120
    assert c.params[8] == 60 / 720
    c.set_max_order(1)
    assert c.params[3] == 10
    assert c.params[4] == 0
    assert c.params[5] == 0
    assert c.params[6] == 0
    assert c.params[7] == 0
    assert c.params[8] == 0
    with pytest.raises(TypeError):
        c.set_max_order("invalid_text")
    with pytest.raises(ValueError):
        c.set_max_order(50)
    c.calculate(20)
    with patch("matplotlib.pyplot.show"):
        c.plot_result()



import sys

sys.path.append("..")

import unittest
from unittest.mock import patch

import numpy as np

from pysprint.core._generator import generator_freq, generator_wave, C_LIGHT
from pysprint import Generator as Generator_from_API


class TestGenerator(unittest.TestCase):
    def setUp(self):
        assert C_LIGHT == 299.792458

    def tearDown(self):
        pass

    def test_errors(self):
        with self.assertRaises(ValueError):
            generator_freq(start=1, stop=2, center=3, delay=0)
        with self.assertRaises(ValueError):
            generator_freq(start=5, stop=2, center=3, delay=0)
        with self.assertRaises(ValueError):
            generator_freq(start=1, stop=-1, center=3, delay=0)
        with self.assertRaises(ValueError):
            generator_freq(start=1, stop=3, center=2, delay=0, pulse_width=-20)
        with self.assertRaises(ValueError):
            generator_wave(start=400, stop=800, center=600, delay=0, resolution=1000)

    def test_freq(self):
        a, b, c, d = generator_freq(1, 2, 1.5, delay=0, include_arms=True)
        np.testing.assert_array_equal(c, d)
        assert len(a) == len(b)
        e, f, _, _ = generator_freq(1, 2, 1.5, delay=0)
        assert len(e) == len(f)

    def test_wave(self):
        a, b, c, d = generator_wave(1, 2, 1.5, delay=0, include_arms=True)
        np.testing.assert_array_equal(c, d)
        assert len(a) == len(b)
        e, f, _, _ = generator_wave(1, 2, 1.5, delay=0)
        assert len(e) == len(f)

    @patch("matplotlib.pyplot.show")
    def test_Generator_from_API(self, mock_show):
        g = Generator_from_API(1, 3, 2, 100)
        g.generate_freq()
        g.phase_graph()
        mock_show.assert_called()


if __name__ == "__main__":
    unittest.main()



import numpy as np
from matplotlib import rcParams
from matplotlib import pyplot as plt

from pysprint.utils import _calc_envelope
from pysprint.mpl_tools.peak import SelectButton


class DraggableEnvelope:

    # max absolute pixel distance to count as a hit
    epsilon = 5

    def __init__(self, x, y, mode="l"):
        plt.ion()
        rcParams["toolbar"] = "toolmanager"
        self.fig, self.ax = plt.subplots()
        self.x = x
        self.y = y
        self.mode = mode
        if self.mode == "l":
            self.envelope, self.y_env, self.loc = _calc_envelope(
                self.y, np.arange(len(self.y)), "l"
            )
            plt.title("Adjust the lower envelope.")
        elif self.mode == "u":
            self.envelope, self.y_env, self.loc = _calc_envelope(
                self.y, np.arange(len(self.y)), "u"
            )
            plt.title("Adjust the upper envelope.")
        else:
            raise ValueError("mode must be u or l.")
        # the active point index
        self._ind = None
        self.x_env = self.x[self.loc]

        (self.basedata,) = self.ax.plot(self.x, self.y)
        (self.lines,) = self.ax.plot(self.x, self.envelope, "r")
        (self.peakplot,) = self.ax.plot(self.x_env, self.y_env, "ko")

        self.fig.canvas.mpl_connect("button_press_event", self.button_press_callback)
        self.fig.canvas.mpl_connect("key_press_event", self.key_press_callback)
        self.fig.canvas.mpl_connect("draw_event", self.draw_callback)
        self.fig.canvas.mpl_connect(
            "button_release_event", self.button_release_callback
        )
        self.fig.canvas.mpl_connect("motion_notify_event", self.motion_notify_callback)

        tm = self.fig.canvas.manager.toolmanager
        tm.add_tool("Toggle recording", SelectButton)

        self.fig.canvas.manager.toolbar.add_tool(
            tm.get_tool("Toggle recording"), "toolgroup"
        )

        self.my_select_button = tm.get_tool("Toggle recording")
        plt.grid()
        plt.show(block=True)

    def button_release_callback(self, event):
        """whenever a mouse button is released"""
        if event.button != 1:
            return
        self._ind = None

    def get_ind_under_point(self, event):
        """
        Get the index of the selected point within the given epsilon tolerance.
        """

        # We use the pixel coordinates, because the axes are usually really
        # differently scaled.
        x, y = self.peakplot.get_data()
        xy_pixels = self.ax.transData.transform(np.vstack([x, y]).T)
        xpix, ypix = xy_pixels.T

        # return the index of the point iff within epsilon distance.
        d = np.hypot(xpix - event.x, ypix - event.y)
        (indseq,) = np.nonzero(d == d.min())
        ind = indseq[0]

        if d[ind] >= self.epsilon:
            ind = None
        return ind

    def button_press_callback(self, event):
        """whenever a mouse button is pressed we get the index"""
        if event.inaxes is None:
            return
        if event.button != 1:
            return
        self._ind = self.get_ind_under_point(event)

    def key_press_callback(self, event):
        """whenever a key is pressed"""
        if not event.inaxes:
            return
        if self.my_select_button.toggled:
            if event.key == "d":
                ind = self.get_ind_under_point(event)
                if ind is not None:
                    self.x_env = np.delete(self.x_env, ind)
                    self.y_env = np.delete(self.y_env, ind)
                    self._interpolate()
                    self.peakplot.set_data(self.x_env, self.y_env)
                    self.lines.set_data(self.x, self.envelope)
            elif event.key == "i":
                self.y_env = np.append(self.y_env, event.ydata)
                self.x_env = np.append(self.x_env, event.xdata)
                self._interpolate()
                self.peakplot.set_data(self.x_env, self.y_env)
                self.lines.set_data(self.x, self.envelope)
            if self.peakplot.stale:
                self.fig.canvas.draw_idle()

    def _interpolate(self):
        """Upon modifying the datapoints we need to sort values and
        recalculate the interpolation."""
        idx = np.argsort(self.x_env)
        self.y_env, self.x_env = self.y_env[idx], self.x_env[idx]
        self.envelope = np.interp(self.x, self.x_env, self.y_env)

    def get_data(self):
        """
        Only returns the y values accordingly
        """
        if self.mode == "l":
            return self.y - self.envelope
        elif self.mode == "u":
            return self.y / self.envelope

    def draw_callback(self, event):
        self.background = self.fig.canvas.copy_from_bbox(self.ax.bbox)
        self.ax.draw_artist(self.peakplot)
        self.ax.draw_artist(self.lines)

    def motion_notify_callback(self, event):
        """on mouse movement we move the selected point"""
        if self._ind is None:
            return
        if event.inaxes is None:
            return
        if event.button != 1:
            return
        if self.my_select_button.toggled:
            x, y = event.xdata, event.ydata
            self.x_env[self._ind], self.y_env[self._ind] = x, y
            self._interpolate()
            self.peakplot.set_data(self.x_env, self.y_env)
            self.lines.set_data(self.x, self.envelope)

            # TODO: make faster drawing with background
            # self.fig.canvas.restore_region(self.background)

            self.ax.draw_artist(self.lines)
            self.ax.draw_artist(self.peakplot)
            self.fig.canvas.blit(self.ax.bbox)
            self.fig.canvas.draw()



import sys
import unittest

sys.modules["lmfit"] = None

import numpy as np

from pysprint.core import _evaluate
from pysprint import Generator, FFTMethod


class TestEvaluateNoLmfit(unittest.TestCase):
    def setUp(self):
        _has_lmfit = False

    def test_cff(self):
        _has_lmfit = False
        a = np.arange(100)
        b = np.arange(100)
        with self.assertRaises(KeyError):
            _evaluate.cff_method(
                a, b, [], [], ref_point=0, p0=[1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
            )

    def test_ffts_primitive(self):
        # adapted from scipy's unittests
        np.random.seed(1534)
        x = np.random.randn(10) + 1j * np.random.randn(10)
        fr, yf = _evaluate.ifft_method(x, x, interpolate=False)
        _, y = _evaluate.fft_method(yf, yf)
        np.testing.assert_allclose(y, x)

    def test_windowing(self):
        _has_lmfit = False
        a, b = np.loadtxt("test_window.txt", unpack=True, delimiter=",")
        y_data = _evaluate.cut_gaussian(a, b, 2.5, 0.2, 6)
        assert len(b) == len(y_data)
        np.testing.assert_almost_equal(y_data[0], 0)
        np.testing.assert_almost_equal(y_data[-1], 0)
        np.testing.assert_almost_equal(np.median(y_data), np.median(b), decimal=2)

    def test_ffts_primitive2(self):
        # adapted from scipy's unittests
        np.random.seed(1534)
        x = np.random.randn(10) + 1j * np.random.randn(10)
        fr, yf = _evaluate.ifft_method(x, x, interpolate=False)
        _, y = _evaluate.fft_method(yf, yf)
        np.testing.assert_allclose(y, x)

    def test_ffts_advanced2(self):
        g = Generator(
            2, 2.8, 2.4, delay=1500, GDD=2000, pulse_width=25, resolution=0.01
        )
        g.generate_freq()
        a, b = g.data
        f = FFTMethod(a, b)
        f.ifft()
        f.window(1500, 2920, plot=False)
        f.apply_window()
        f.fft()
        d, _, _ = f.calculate(order=2, reference_point=2.4)
        np.testing.assert_array_almost_equal(d, [1500.01, 1999.79], decimal=2)

    def test_ffts_advanced1(self):
        g = Generator(2, 2.8, 2.4, delay=1500, GD=200, pulse_width=25, resolution=0.01)
        g.generate_freq()
        a, b = g.data
        f = FFTMethod(a, b)
        f.ifft()
        f.window(1700, 3300, plot=False)
        f.apply_window()
        f.fft()
        d, _, _ = f.calculate(order=1, reference_point=2.4)
        np.testing.assert_array_almost_equal(d, [1699.99], decimal=2)

    def test_ffts_advanced3(self):
        g = Generator(
            2, 2.8, 2.4, delay=1500, TOD=40000, pulse_width=25, resolution=0.01
        )
        g.generate_freq()
        a, b = g.data
        f = FFTMethod(a, b)
        f.ifft()
        f.window(2500, 4830, window_order=12, plot=False)
        f.apply_window()
        f.fft()
        d, _, _ = f.calculate(order=3, reference_point=2.4)
        np.testing.assert_array_almost_equal(d, [1500.03, 0.03, 39996.60], decimal=2)

    def test_ffts_advanced4(self):
        g = Generator(
            2,
            2.8,
            2.4,
            delay=1500,
            GDD=2000,
            FOD=-100000,
            pulse_width=25,
            resolution=0.01,
        )
        g.generate_freq()
        a, b = g.data
        f = FFTMethod(a, b)
        f.ifft()
        f.window(1500, 1490, window_order=8, plot=False)
        f.apply_window()
        f.fft()
        d, _, _ = f.calculate(order=4, reference_point=2.4)
        np.testing.assert_array_almost_equal(
            d, [1500.00, 1999.95, -0.21, -99995.00], decimal=1
        )

    def test_ffts_advanced5(self):
        g = Generator(
            2, 2.8, 2.4, delay=1500, QOD=900000, pulse_width=25, resolution=0.01,
        )
        g.generate_freq()
        a, b = g.data
        f = FFTMethod(a, b)
        f.ifft()
        f.window(1600, 2950, window_order=12, plot=False)
        f.apply_window()
        f.fft()
        d, _, _ = f.calculate(order=5, reference_point=2.4)
        np.testing.assert_array_almost_equal(
            d, [1499.9646, -0.1429, 7.893, 16.075, 898919.85], decimal=1
        )


if __name__ == "__main__":
    unittest.main()



import warnings
from contextlib import contextmanager

import matplotlib  # noqa
import matplotlib.pyplot as plt  # noqa

import pysprint.core.init_config
from pysprint.utils.misc import run_from_ipython

warnings.filterwarnings("ignore", message="invalid value encountered in sqrt")
warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")
warnings.filterwarnings("ignore", message="Casting complex values to real discards the imaginary part")

__author__ = "Leéh Péter"

from .core import *
from .utils import print_info

default_backend = "Qt5Agg" if getattr(matplotlib, "__version__", None) != "3.3.1" else "TkAgg"


@contextmanager
def interactive(backend=default_backend, figsize=(15, 5)):
    """
    Context manager to temporarily change the matplotlib
    backend to ensure interactive figures are rendered
    correctly.

    Parameters
    ----------
    backend : string, optional
        The matplotlib backend to use. Default is "Qt5Agg".
    figsize : tuple, optional
        The figure size to use. Default is (15, 5).
    """
    plt.ion()
    matplotlib.rcParams["figure.figsize"] = figsize
    original_backend = plt.get_backend()
    try:
        plt.switch_backend(backend)
        yield
    except (AttributeError, ImportError, ModuleNotFoundError) as err:
        raise ValueError(
            f"Couldn't set backend {backend}, you should manually "
            "change to an appropriate GUI backend. (Matplotlib 3.3.1 "
            "is broken. In that case use backend='TkAgg')."
        ) from err
    finally:
        plt.switch_backend(original_backend)


def set_interactive(backend=default_backend, figsize=(15, 5)):
    """
    Set the backend for matplotlib permanently.

    Parameters
    ----------
    backend : string, optional
        The matplotlib backend to use. Default is "Qt5Agg".
    figsize : tuple, optional
        The figure size to use. Default is (15, 5).
    """
    plt.ion()
    matplotlib.rcParams["figure.figsize"] = figsize
    try:
        plt.switch_backend(backend)
    except (AttributeError, ImportError, ModuleNotFoundError) as err:
        raise ValueError(
            f"Couldn't set backend {backend}, you should manually "
            "change to an appropriate GUI backend. (Matplotlib 3.3.1 "
            "is broken. In that case use backend='TkAgg')."
        ) from err

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import sys

if sys.version_info[:2] < (3, 6):
    raise RuntimeError("Python version >= 3.6 required.")

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

import versioneer

setup(
    name="pysprint",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Péter Leéh",
    author_email="leeh123peter@gmail.com",
    description="Spectrally refined interferometry for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Ptrskay3/PySprint",
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    install_requires=["numpy>=1.16.6", "scipy", "matplotlib", "pandas", "Jinja2"],
    extras_require={"optional": ["numba", "lmfit", "pytest", "dask"]},
    entry_points={
        'console_scripts': [
            'pysprint = pysprint.templates.build:main',
        ],
    }
)



import types
from weakref import WeakSet

from pysprint.utils.decorators import pprint_disp

__all__ = ["_DatasetBase", "C_LIGHT"]

C_LIGHT = 299.792458


class _DatasetBase(type):
    """
    Base metaclass that defines the behaviour of any interferogram
    class, and also this sets up the registry.

    Set autoprinting if subclass implements calculate-like method.
    """
    def __init__(cls, name, bases, attrs):
        super(_DatasetBase, cls).__init__(name, bases, attrs)
        cls._instances = WeakSet()

    def __call__(cls, *args, **kwargs):
        inst = super(_DatasetBase, cls).__call__(*args, **kwargs)
        cls._instances.add(inst)
        return inst

    def _get_instances(cls, recursive=False):
        instances = list(cls._instances)
        if recursive:
            for child in cls.__subclasses__():
                instances += child._get_instances(recursive=recursive)
        return list(set(instances))

    def __new__(cls, name, bases, attrs):
        for attr_name, attr_value in attrs.items():
            if isinstance(attr_value, types.FunctionType):
                if attr_name == "calculate" or attr_name.startswith("calculate"):
                    attrs[attr_name] = pprint_disp(attr_value)
            else:
                if attr_name == "calculate" or attr_name.startswith("calculate"):
                    if isinstance(attr_value, staticmethod):
                        attrs[attr_name] = staticmethod(pprint_disp(attr_value.__func__))

        return super(_DatasetBase, cls).__new__(cls, name, bases, attrs)



import warnings

import numpy as np
from scipy.signal import find_peaks

from pysprint.config import _get_config_value


def signaltonoise(a, axis=0, ddof=0):
    """
    Reimplementation of the deprecated
    signaltonoise function from scipy.
    """
    a = np.asanyarray(a)
    m = a.mean(axis)
    sd = a.std(axis=axis, ddof=ddof)
    return np.where(sd == 0, 0, m / sd)


def find_roi(x, y):
    """
    Find the region of interest in FFT data.
    """
    y = np.abs(y)
    idx = np.where(x > 0)
    return x[idx], y[idx]


def find_center(x, y, n_largest=5, return_multiple=None):
    """
    Find the center of the peak in FFT'd data.
    """
    peaks, props = find_peaks(y, prominence=0.001, height=np.max(y) / 100)

    if n_largest > len(props["prominences"]):
        n_largest = len(props["prominences"])

    # find the most outlying peaks from the noise
    ind1 = np.argpartition(props["prominences"], -n_largest)[-n_largest:]
    # find the highest peaks by value
    ind2 = np.argpartition(props["peak_heights"], -n_largest)[-n_largest:]

    ind = np.unique(np.concatenate((ind1, ind2)))

    peaks = peaks[ind]
    y_prob_density = np.exp((-((x - np.max(x) / 2.5) ** 2)) / (1000 * np.max(x)))
    _x, _y, _y_ = x[peaks], y_prob_density[peaks], y[peaks]
    # weighted with the prob density function above from origin
    try:
        residx = np.argmax(_x * _y * _y_)
    except ValueError as err:
        msg = ValueError(
            "Probably you need to set bigger window FWHM. "
            "After IFFT, the center of the peak could not be determined."
        )
        raise msg from err

    if return_multiple:
        try:
            N = int(return_multiple)
        except ValueError as err:
            raise ValueError("Must return integer number of peaks.") from err
        xval, yval = x[peaks], y[peaks]
        if len(xval) < N:
            N = len(xval)
        idx_ = np.argsort(xval)
        retx, rety = xval[idx_], yval[idx_]
        return retx[:N], rety[:N]

    return _x[residx], _y_[residx]


def _ensure_window_at_origin(center, fwhm, order, peak_center_height, tol=1e-3):
    """
    Ensure that the gaussian window of given parameters is
    not crossing zero with bigger value than the desired tolerance.
    """
    std = fwhm / (2 * (np.log(2) ** (1 / order)))
    val = np.exp(-((0 - center) ** order) / (std ** order))
    return val < peak_center_height * tol, val


def predict_fwhm(x, y, center, peak_center_height, prefer_high_order=True, tol=1e-3):
    """
    Predicts ideal fwhm for Gaussian windows. Still needs work to be reliable.
    """
    if np.iscomplexobj(y):
        y = np.abs(y)

    N = len(x)

    peak_position = center / N

    # explicitly prefer higher order windows
    # if the peak is too close to the origin

    if peak_position < 0.2:
        warnings.warn(
            "The peak is too close to the origin, manual control is advised.",
            UserWarning,
        )
        prefer_high_order = True

    if prefer_high_order:
        order_choices = (8, 10, 12)
    else:
        order_choices = (2, 4, 6)

    sgnl = signaltonoise(y)
    if peak_position < 0.6:
        if sgnl > 0.1:
            center += N * 0.05 + (sgnl - 0.1) * N * 2
            window_size = 0.75 * center * 2 + (sgnl - 0.1) * N * 0.2
        else:
            center += N * 0.05
            window_size = 0.75 * center * 2
    else:
        if sgnl > 0.1:
            center += N * 0.05 + (sgnl - 0.1) * N * 2
            window_size = 0.5 * center * 2 + (sgnl - 0.1) * N * 0.2
        else:
            center += N * 0.05
            window_size = 0.5 * center * 2
    try:
        order = order_choices[int(2 - (3 * peak_position // 2))]
    except IndexError:
        order = order_choices[0]

    if _ensure_window_at_origin(
            center, window_size, order, peak_center_height, tol=tol
    )[0]:
        return center, window_size, order
    else:
        val = _ensure_window_at_origin(
            center, window_size, order, peak_center_height, tol=tol
        )[1]
        warnings.warn(
            "The window is bigger at the origin than the desired tolerance. "
            f"Actual:{val:.4e}, Desired:{tol * peak_center_height:.4e}"
        )
        return center, window_size, order + 2


def _run(
        ifg, skip_domain_check=False, show_graph=True, usenifft=False,
):
    precision = _get_config_value("precision")
    print("Interferogram received.")
    if ifg.probably_wavelength and not skip_domain_check:
        print(
            "Probably in wavelength domain, changing to frequency...",
            end="",
            flush=True,
        )
        ifg.chdomain()
        print("Done")
    if usenifft:
        pprint = " using NUIFFT algorithm..."
    else:
        pprint = "..."
    print(f"Applying IFFT{pprint}", end="", flush=True)
    ifg.ifft(usenifft=usenifft)
    print("Done")
    x, y = find_roi(ifg.x, ifg.y)
    a, b = find_center(x, y)
    print("Acquiring gaussian window parameters...", end="", flush=True)
    print("Done")
    c, ws, _order = predict_fwhm(x, y, center=a, peak_center_height=b)
    print(
        f"A {_order} order gaussian window centered at "
        f"{c:.{precision}f} fs with FWHM {ws:.{precision}f} fs was applied."
    )
    ifg.window(at=c, fwhm=ws, window_order=_order, plot=show_graph)
    ifg.apply_window()
    print("Applying FFT...", end="", flush=True)
    ifg.fft()
    print("Done")
    print("Calculating...")



import os

import numpy as np
from scipy.interpolate import interp1d
import scipy.stats as st

from pysprint.config import _get_config_value

__all__ = [
    "_unpack_lmfit",
    "find_nearest",
    "_handle_input",
    "_fourier_interpolate",
    "_between",
    "_get_closest",
    "run_from_ipython",
    "_calc_envelope",
    "measurement",
    "_maybe_increase_before_cwt",
    "pad_with_trailing_zeros",
    "pprint_math_or_default",
]


def pprint_math_or_default(s):
    try:
        from IPython.display import display, Math
        display(
            Math(
                s
            )
        )
    except ImportError:
        print(s)


def _maybe_increase_before_cwt(y, tolerance=0.05):
    y = np.asarray(y)
    value = y.min()
    if value - tolerance >= 0:
        return False
    if value < 0 < np.abs(value) - tolerance:
        return True
    return True


def _calc_envelope(x, ind, mode="u"):
    """
    https://stackoverflow.com/a/39662343/11751294
    """
    x_abs = np.abs(x)
    if mode == "u":
        loc = np.where(np.diff(np.sign(np.diff(x_abs))) < 0)[0] + 1
    elif mode == "l":
        loc = np.where(np.diff(np.sign(np.diff(x_abs))) > 0)[0] + 1
    else:
        raise ValueError("mode must be u or l.")
    peak = x_abs[loc]
    envelope = np.interp(ind, loc, peak)
    return envelope, peak, loc


def run_from_ipython():
    """
    Detect if code is run inside Jupyter or maybe Spyder.
    """
    try:
        __IPYTHON__  # noqa
        if any("SPYDER" in name for name in os.environ):
            return False
        return True
    except NameError:
        return False


def _get_closest(x_val, y_val, x_array, y_array):
    """
    Get the closest 2D point in array.
    """
    idx = np.argmin((np.hypot(x_array - x_val, y_array - y_val)))
    value = x_array[idx]
    return value, y_array[idx], idx


def _between(val, except_around):
    if except_around is None:
        return False
    elif len(except_around) != 2:
        raise ValueError(
            f"Invalid interval. Try [start, end] instead of {except_around}"
        )
    else:
        lower = float(min(except_around))
        upper = float(max(except_around))
    if upper >= val >= lower:
        return True
    return False


def _unpack_lmfit(r):
    dispersion, dispersion_std = [], []
    for name, par in r:
        dispersion.append(par.value)
        dispersion_std.append(par.stderr)
    return dispersion, dispersion_std


def measurement(array, confidence=0.95, silent=False):
    """
    Give the measurement results with condifence interval
    assuming the standard deviation is unknown.

    Parameters
    ----------

    array : ndarray
        The array containing the measured values

    confidence : float, optional
        The desired confidence level. Must be _between 0 and 1.

    silent : bool, optional
        Whether to print results immediately. Default is `False`.

    Returns
    -------

    mean: float
        The mean of the given array

    conf: tuple-like (interval)
        The confidence interval

    Examples
    ---------
    >>> import numpy as np
    >>> from pysprint.utils import measurement
    >>> a = np.array([123.783, 121.846, 122.248, 125.139, 122.569])
    >>> mean, interval = measurement(a, 0.99)
    123.117000 ± 2.763022
    >>> mean
    123.117
    >>> interval
    (120.35397798230359, 125.88002201769642)

    Notes
    -----

    I decided to print the results immediately, because people often don't use
    it for further code. Of course, they are also returned if needed.
    """
    precision = _get_config_value("precision")
    mean = np.mean(array)
    conf = st.t.interval(confidence, len(array) - 1, loc=mean, scale=st.sem(array))
    if not silent:
        pprint_math_or_default(
            f"{mean:.{precision}f} ± {(mean - conf[0]):.{precision}f}"
        )
    return mean, conf


def find_nearest(array, value):
    """
    Find the nearest element in array to value.

    Parameters
    ----------
    array : np.ndarray-like
        The array to search in.
    value : float
        The value to search.

    Returns
    -------
    value : float
        The closest value in array.
    idx : int
        The index of the closest element.
    """
    array = np.asarray(array)
    idx = (np.abs(value - array)).argmin()
    return array[idx], idx


def _handle_input(x, y, ref, sam):
    """
    Instead of handling the inputs in every function, there is this private method.

    Parameters
    ----------

    x: array-like
    x-axis data

    y: array-like
    y-axis data

    ref, sam: array-like
    reference and sample arm spectrum evaluated at x

    Returns
    -------
    x: array-like
    unchanged x data

    Ydata: array-like
    the transformed y data

    """
    if len(x) == 0:
        raise ValueError("No values for x.")

    if len(y) == 0:
        raise ValueError("No values for y.")

    if (len(x) > 0) and (len(ref) > 0) and (len(sam) > 0):
        y_data = (y - ref - sam) / (2 * np.sqrt(ref * sam))
    elif (len(ref) == 0) or (len(sam) == 0):
        y_data = y
    else:
        raise TypeError("Input types are wrong.\n")
    return np.asarray(x), np.asarray(y_data)


def _fourier_interpolate(x, y):
    """ Simple linear interpolation for FFTs"""
    xs = np.linspace(x[0], x[-1], len(x))
    intp = interp1d(x, y, kind="linear", fill_value="extrapolate")
    ys = intp(xs)
    return xs, ys


def pad_with_trailing_zeros(array, shape):
    """
    Pad an array with trailing zeros to be the desired shape
    """
    c = shape - len(array)
    if c < 1:
        return np.asarray(array)
    return np.pad(array, pad_width=(0, c), mode="constant")



import types
import warnings
from collections.abc import Iterable
from inspect import getfullargspec

import numpy as np


class _DatasetApply:
    """
    Helper class to apply function to
    `pysprint.core.bases.dataset.Dataset` objects.
    """
    def __init__(
            self,
            obj,
            func,
            axis=None,
            args=None,
            kwargs=None
    ):

        self.obj = obj
        self.args = args or ()
        self.kwargs = kwargs or {}

        self.f = func
        self.axis = axis

        if self.axis == "x" or self.axis == 0:
            self.target = "x"
        elif self.axis == "y" or self.axis == 1:
            self.target = "y"
        else:
            raise ValueError("Axis must be 'x', 'y', '0' or '1'.")
        self.shape = len(getattr(self.obj, self.target))

    def perform(self):
        """
        Apply the specified function.
        """
        if isinstance(self.f, str):
            func = getattr(self.obj, self.f)
            sig = getfullargspec(func)
            if "axis" in sig.args:
                self.kwargs["axis"] = self.axis
            # Let's assume we don't mess up the shape internally
            func(*self.args, **self.kwargs)
            return self.obj  # we need to return this because of `inplacify` deco.

        elif isinstance(self.f, np.ufunc):
            target = getattr(self.obj, self.target)

            retval = self.f(target, *self.args, **self.kwargs)
            value = self._validate(retval)

            setattr(self.obj, self.target, value)
            if self.target == "y":
                setattr(self.obj, "y_norm", value)
            return value

        elif isinstance(self.f, types.FunctionType):
            sig = getfullargspec(self.f)
            if "axis" in sig.args:
                self.kwargs["axis"] = self.axis
            # we can safely vectorize it here
            self.f = np.vectorize(self.f)
            target = getattr(self.obj, self.target)
            retval = self.f(target, *self.args, **self.kwargs)
            value = self._validate(retval)
            setattr(self.obj, self.target, value)
            if self.target == "y":
                setattr(self.obj, "y_norm", value)
            return value

    def _validate(self, val):

        if isinstance(val, (Iterable, np.ndarray)):
            val = np.asarray(val, dtype=np.float64)

            if val.ndim != 1:
                val = np.concatenate(val).ravel()
                warnings.warn("Function return value was flattened.")

            if len(val) != len(np.unique(val)):
                if len(np.unique(val)) == self.shape:
                    return val
                else:
                    if self.target == "x":
                        raise ValueError(
                            f"Function returned duplicated values which is not allowed when"
                            " modifying the x axis. After filtering to unique values "
                            f"a {len(np.unique(val))}-length array was produced, "
                            f"but {self.shape} was expected."
                        )
                    return val

            if len(val) != self.shape:
                retval = self._broadcast(val)
                return retval
            return val
        else:
            raise TypeError("Function should return a number or Iterable type.")

    def _broadcast(self, val):
        if len(val) > self.shape:
            return val[:self.shape]
        elif len(val) < self.shape:
            if not self.shape % len(val) == 0:
                raise ValueError("Cannot broadcast safely to the desired shape.")
            else:
                return np.repeat(val, (self.shape % len(val)))



import numpy as np


def longest_common_subsequence(x1, y1, x2, y2, tol=None):
    """
    Given two datasets with x-y values, find the longest common
    subsequence of them including a small threshold which might be
    present due to numerical errors.
    This function is mainly used when two datasets's y values need
    to be multiplied together, but their domains are slightly off.

    Parameters
    ----------
    x1 : np.ndarray-like
        The x values for the original array
    y1 : np.ndarray-like
        The y values for the original array
    x2 : np.ndarray-like
        The x values for the second array
    y2 : np.ndarray-like
        The y values for the second array
    tol : float, optional
        The tolerance which determines how big difference is allowed
        _between x values to interpret them as the same datapoint.

    Returns
    -------
    longest_x1: np.ndarray
        The x values of longest common subsequence in x1.
    longest_y1: np.ndarray
        The y values of longest common subsequence in y1.
    longest_x2: np.ndarray
        The x values of longest common subsequence in x2.
    longest_y2: np.ndarray
        The y values of longest common subsequence in y2.
    """
    # sort both datasets first to maintain order
    idx1 = np.argsort(x1)
    x1, y1 = x1[idx1], y1[idx1]
    idx2 = np.argsort(x2)
    x2, y2 = x2[idx2], y2[idx2]

    if tol is None:
        tol = np.max(np.diff(x2))

    dist = np.abs(x1 - x2[:, None])
    i1 = np.argmin(dist, axis=1)
    i2 = np.flatnonzero(dist[np.arange(x2.size), i1] < tol)
    i1 = i1[i2]

    mask = (np.diff(i1) == 1) & (np.diff(i2) == 1)

    # smear the mask to include both endpoints
    mask = np.r_[False, mask] | np.r_[mask, False]

    # pad the mask to ensure proper indexing and find the changeover points
    locs = np.diff(np.r_[False, mask, False])
    inds = np.flatnonzero(locs)
    lengths = inds[1::2] - inds[::2]

    k = np.argmax(lengths)
    start = inds[2 * k]
    stop = inds[2 * k + 1]

    longest_x1 = x1[i1[start:stop]]
    longest_y1 = y1[i1[start:stop]]
    longest_x2 = x2[i2[start:stop]]
    longest_y2 = y2[i2[start:stop]]

    return longest_x1, longest_y1, longest_x2, longest_y2



import sys
import warnings
from inspect import isfunction

import numpy as np
import matplotlib.pyplot as plt

from pysprint.config import _get_config_value
from pysprint.core.methods.fftmethod import FFTMethod
from pysprint.core.phase import Phase
from pysprint.core._evaluate import gaussian_window
from pysprint.core._fft_tools import find_roi
from pysprint.core._fft_tools import find_center
from pysprint.utils.decorators import _mutually_exclusive_args
from pysprint.utils.decorators import _lazy_property
from pysprint.utils.decorators import inplacify
from pysprint.utils import NotCalculatedException
from pysprint.utils import PySprintWarning
from pysprint.utils.misc import find_nearest

try:
    from dask import delayed, compute
    from dask.diagnostics import ProgressBar
    CAN_PARALLELIZE = True
except ImportError:
    CAN_PARALLELIZE = False

    def delayed(func=None, *args, **kwargs):
        """Define delayed as identity deco."""
        if isfunction(func):
            return func


class Window:
    """
    Basic class that implements functionality related to Gaussian
    windows with caching the y values.
    """

    def __init__(self, x, center, fwhm, order=2):
        self.x = x
        self.center = center
        self.fwhm = fwhm
        self.order = order

    @_lazy_property
    def y(self):
        """
        The y values of the given window. It's a "lazy_property".
        """
        return gaussian_window(self.x, self.center, self.fwhm, self.order)

    @classmethod
    def from_std(cls, x, center, std, order=2):
        """
        Build the Gaussian window from standard deviation instead of fwhm.
        """
        _fwhm = std * 2 * np.log(2) ** (1 / order)
        return cls(x, center, _fwhm, order)

    def __str__(self):
        precision = _get_config_value("precision")
        return f"Window(center={self.center:.{precision}f}, fwhm={self.fwhm}, order={self.order})"

    def plot(self, ax=None, scalefactor=1, zorder=90, **kwargs):
        """
        Plot the window.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            The axis to plot on. If not given, plot on the last axis.
        scalefactor : float, optional
            Number describing how much a given window should be scaled up ONLY
            for visibility.
        zorder : float, optional
            The drawing order of artists is determined by their zorder attribute, which is
            a floating point number. Artists with higher zorder are drawn on top. You can
            change the order for individual artists by setting their zorder. The default
            value depends on the type of the Artist.
        """
        if ax is None:
            ax = plt
        ax.plot(self.x, self.y * scalefactor, zorder=zorder, **kwargs)


class WFTMethod(FFTMethod):
    """Basic interface for Windowed Fourier Transform Method."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.window_seq = {}
        self.found_centers = {}
        self.GD = None
        self.cachedlen = 0
        self.X_cont = np.array([])
        self.Y_cont = np.array([])
        self.Z_cont = np.array([])
        self.fastmath = True
        self.errorcounter = 0

    @inplacify
    @_mutually_exclusive_args("std", "fwhm")
    def add_window(self, center, std=None, fwhm=None, order=2):
        """
        Add a Gaussian window to the interferogram.

        Parameters
        ----------
        center : float
            The center of the Gaussian window.
        std : float, optional
            The standard deviation of the Gaussian window
            in units of the x axis of the interferogram.
            You must specify exactly one of std and fwhm.
        fwhm : float, optional
            The full width at half max of the Gaussian window
            in units of the x axis of the interferogram.
            You must specify exactly one of std and fwhm.
        order : int, optional
            The order of Gaussian window. Must be even.
            The default is 2.
        """
        if not np.min(self.x) <= center <= np.max(self.x):
            raise ValueError(
                f"Cannot add window at {center}, because "
                f"it is out of the dataset's range (from {np.min(self.x):.3f} to {np.max(self.x):.3f})."
            )
        if std:
            window = Window.from_std(
                self.x, center=center, std=std, order=order
            )
        else:
            window = Window(self.x, center=center, fwhm=fwhm, order=order)
        self.window_seq[center] = window
        return self

    @property
    def windows(self):
        return self.window_seq

    @property
    def centers(self):
        return self.window_seq.keys()

    @inplacify
    @_mutually_exclusive_args("std", "fwhm")
    def add_window_generic(self, array, std=None, fwhm=None, order=2):
        """
        Build a window sequence of given parameters with centers
        specified with ``array`` argument.

        Parameters
        ----------
        array : list, np.ndarray
            The array containing the centers of windows.
        std : float, optional
            The standard deviation of the Gaussian window
            in units of the x axis of the interferogram.
            You must specify exactly one of std and fwhm.
        fwhm : float, optional
            The full width at half max of the Gaussian window
            in units of the x axis of the interferogram.
            You must specify exactly one of std and fwhm.
        order : int, optional
            The order of Gaussian window. Must be even.
            The default is 2.
        """
        if not isinstance(array, (list, np.ndarray)):
            raise TypeError("Expected list-like as ``array``.")
        for center in array:
            if std:
                self.add_window(center=center, std=std, order=order)
            else:
                self.add_window(center=center, fwhm=fwhm, order=order)
        return self

    @inplacify
    @_mutually_exclusive_args("std", "fwhm")
    def add_window_arange(
        self, start, stop, step, std=None, fwhm=None, order=2
    ):
        """
        Build a window sequence of given parameters to apply on ifg.
        Works similar to numpy.arange.

        Parameters
        ----------
        start : float
            The start of the centers.
        stop : float
            The end value of the center
        step : float
            The step value to increment center.
        std : float, optional
            The standard deviation of the Gaussian window
            in units of the x axis of the interferogram.
            You must specify exactly one of std and fwhm.
        fwhm : float, optional
            The full width at half max of the Gaussian window
            in units of the x axis of the interferogram.
            You must specify exactly one of std and fwhm.
        order : int, optional
            The order of Gaussian window. Must be even.
            The default is 2.
        """
        arr = np.arange(start, stop, step)
        for cent in arr:
            if std:
                self.add_window(center=cent, std=std, order=order)
            else:
                self.add_window(center=cent, fwhm=fwhm, order=order)
        return self

    @inplacify
    @_mutually_exclusive_args("std", "fwhm")
    def add_window_linspace(
        self, start, stop, num, std=None, fwhm=None, order=2
    ):
        """
        Build a window sequence of given parameters to apply on ifg.
        Works similar to numpy.linspace.

        Parameters
        ----------
        start : float
            The start of the centers.
        stop : float
            The end value of the center
        num : float
            The number of Gaussian windows.
        std : float, optional
            The standard deviation of the Gaussian window
            in units of the x axis of the interferogram.
            You must specify exactly one of std and fwhm.
        fwhm : float, optional
            The full width at half max of the Gaussian window
            in units of the x axis of the interferogram.
            You must specify exactly one of std and fwhm.
        order : int, optional
            The order of Gaussian window. Must be even.
            The default is 2.
        """
        arr = np.linspace(start, stop, num)
        for cent in arr:
            if std:
                self.add_window(center=cent, std=std, order=order)
            else:
                self.add_window(center=cent, fwhm=fwhm, order=order)
        return self

    @inplacify
    @_mutually_exclusive_args("std", "fwhm")
    def add_window_geomspace(
        self, start, stop, num, std=None, fwhm=None, order=2
    ):
        """
        Build a window sequence of given parameters to apply on ifg.
        Works similar to numpy.geomspace.

        Parameters
        ----------
        start : float
            The start of the centers.
        stop : float
            The end value of the center
        num : float
            The number of Gaussian windows.
        std : float, optional
            The standard deviation of the Gaussian window
            in units of the x axis of the interferogram.
            You must specify exactly one of std and fwhm.
        fwhm : float, optional
            The full width at half max of the Gaussian window
            in units of the x axis of the interferogram.
            You must specify exactly one of std and fwhm.
        order : int, optional
            The order of Gaussian window. Must be even.
            The default is 2.
        """
        arr = np.geomspace(start, stop, num)
        for cent in arr:
            if std:
                self.add_window(center=cent, std=std, order=order)
            else:
                self.add_window(center=cent, fwhm=fwhm, order=order)
        return self

    def view_windows(self, ax=None, maxsize=80, **kwargs):
        """
        Gives a rough view of the different windows along with the ifg.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            An axis to draw the plot on. If not given, it will plot
            of the last used axis.
        maxsize : int, optional
            The maximum number of Gaussian windows to display on plot.
            Default is 80, but be aware that setting a high value can
            drastically reduce performance.
        kwargs : dict, optional
            Additional keyword arguments to pass to plot function.
        """
        winlen = len(self.window_seq)
        if maxsize != 0:
            ratio = winlen // maxsize
            if winlen > maxsize:
                warnings.warn(
                    "Image seems crowded, displaying only a subsample of the given windows.",
                    PySprintWarning
                )
                for i, (_, val) in enumerate(self.window_seq.items()):
                    if i % ratio == 0:
                        val.plot(ax=ax, scalefactor=np.max(self.y) * .75, **kwargs)
            else:
                for _, val in self.window_seq.items():
                    val.plot(ax=ax, scalefactor=np.max(self.y) * .75, **kwargs)
        self.plot(ax=ax)

    @inplacify
    def remove_all_windows(self):
        """
        Remove all the Gaussian windows.
        """
        self.window_seq.clear()
        return self

    @inplacify
    def reset_state(self):
        """
        Reset the object's state fully: delete all the
        calculated GD, caches, heatmaps and window sequences.
        """
        self.remove_all_windows()
        self.found_centers.clear()
        self.X_cont = np.array([])
        self.Y_cont = np.array([])
        self.Z_cont = np.array([])
        self.GD = None
        self.cachedlen = 0
        self.fastmath = True
        return self

    @inplacify
    def remove_window_at(self, center):
        """
        Removes a window at center.

        Parameters
        ----------
        center : float
            The center of the window to remove.
            Raises ValueError if there is not such window.
        """
        if center not in self.window_seq.keys():
            c = find_nearest(
                np.fromiter(self.window_seq.keys(), dtype=float), center
            )
            raise ValueError(
                f"There is no window with center {center}. "
                f"Did you mean {c[0]}?"
            )
        self.window_seq.pop(center, None)
        return self

    @inplacify
    def remove_window_interval(self, start, stop):
        """
        Remove window interval inclusively.

        Parameters
        ----------
        start : float
            The start value of the interval.
        stop : float
            The stop value of the interval.
        """
        wins = np.fromiter(self.window_seq.keys(), dtype=float)
        mask = wins[(wins <= stop) & (wins >= start)]
        for center in mask:
            self.window_seq.pop(center, None)
        return self

    def calculate(
            self,
            reference_point,
            order,
            show_graph=False,
            silent=False,
            force_recalculate=False,
            fastmath=True,
            usenifft=False,
            parallel=False,
            errors="ignore"
    ):
        """
        Calculates the dispersion.

        Parameters
        ----------
        reference_point : float
            The reference point.
        order : int
            The dispersion order to look for. Must be in [2, 6].
        show_graph : bool, optional
            Whether to show the GD graph on complete. Default is False.
        silent : bool, optional
            Whether to print progressbar. By default it will print.
        force_recalculate : bool, optional
            Force to recalculate the GD graph not only the curve fitting.
            Default is False.
        fastmath : bool, optional
            Whether to build additional arrays to display heatmap.
            Default is True.
        usenifft : bool, optional
            Whether to use Non-unfirom FFT when calculating GD.
            Default is False. **Not stable.**
        parallel : bool, optional
            Whether to use parallel computation. Only availabe if `Dask`
            is installed. The speedup is about 50-70%. Default is False.
        errors : str, optional
            Whether to raise an error is the algorithm couldn't find the
            center of the peak.

        Raises
        ------
        ValueError, if no window sequence is added to the interferogram.
        ValueError, if order is 1.
        ModuleNotFoundError, if `Dask` is not available when using parallel=True.
        """
        if len(self.window_seq) == 0:
            raise ValueError("Before calculating a window sequence must be set.")

        if self.cachedlen != len(self.window_seq) or fastmath != self.fastmath:
            force_recalculate = True
        self.fastmath = fastmath
        if force_recalculate:
            self.found_centers.clear()
            self.build_GD(
                silent=silent, fastmath=fastmath, usenifft=usenifft, parallel=parallel, errors=errors
            )
        if self.GD is None:
            self.build_GD(
                silent=silent, fastmath=fastmath, usenifft=usenifft, parallel=parallel, errors=errors
            )

        self.cachedlen = len(self.window_seq)

        if order == 1:
            raise ValueError("Cannot fit constant function to data. Order must be in [2, 6].")

        d, ds, fr = self.GD._fit(
            reference_point=reference_point, order=order
        )
        if show_graph:
            self.GD.plot()
        return d, ds, fr

    def build_GD(self, silent=False, fastmath=True, usenifft=False, parallel=False, errors="ignore"):
        """
        Build the GD.

        Parameters
        ----------
        silent : bool, optional
            Whether to print progressbar. By default it will print.
        fastmath : bool, optional
            Whether to build additional arrays to display heatmap.
            Default is True.
        usenifft : bool, optional
            Whether to use Non-unfirom FFT when calculating GD.
            Default is False. **Not stable.**
        parallel : bool, optional
            Whether to use parallel computation. Only availabe if `Dask`
            is installed. The speedup is about 50-70%. Default is False.
        errors : str, optional
            Whether to raise an error is the algorithm couldn't find the
            center of the peak.

        Returns
        -------
        GD : pysprint.core.phase.Phase
            The phase object with `GD_mode=True`. See its docstring for more info.
        """
        if parallel:

            if not CAN_PARALLELIZE:
                raise ModuleNotFoundError(
                    "Module `dask` not found. Please install it in order to use parallelism."
                )

            else:
                self.fastmath = fastmath
                self._apply_window_seq_parallel(fastmath=fastmath, usenifft=usenifft, errors=errors)

                if not silent:
                    with ProgressBar():
                        computed = compute(*self.found_centers.values())

                else:
                    computed = compute(*self.found_centers.values())

                cleaned_delays = [
                    k for i, k in enumerate(self.found_centers.keys()) if computed[i] is not None
                ]
                delay = np.fromiter(cleaned_delays, dtype=float)

                omega = np.fromiter([c for c in computed if c is not None], dtype=float)

                if not silent:
                    print(f"Errors found: {len(self.window_seq) - sum(1 for _ in filter(None.__ne__, computed))}")

        else:
            self.fastmath = fastmath
            self._apply_window_sequence(silent=silent, fastmath=fastmath, usenifft=usenifft)
            self._clean_centers()

            delay = np.fromiter(self.found_centers.keys(), dtype=float)
            omega = np.fromiter(self.found_centers.values(), dtype=float)

        self.GD = Phase(delay, omega, GD_mode=True)
        return self.GD

    def _predict_ideal_window_fwhm(self):
        pass

    def _apply_window_sequence(
            self, silent=False, fastmath=True, usenifft=False, errors="ignore"
    ):
        winlen = len(self.window_seq)
        self.errorcounter = 0
        if not fastmath:
            # here we setup the shape for the Z array because
            # it is much faster than using np.append in every iteration
            _x, _y, _, _ = self._safe_cast()
            _obj = FFTMethod(_x, _y)
            _obj.ifft(usenifft=usenifft)
            x, y = find_roi(_obj.x, _obj.y)
            self.Y_cont = np.array(x)
            yshape = y.size
            xshape = len(self.window_seq)
            self.Z_cont = np.empty(shape=(yshape, xshape))

        for idx, (_center, _window) in enumerate(self.window_seq.items()):
            _x, _y, _, _ = self._safe_cast()
            _obj = FFTMethod(_x, _y)
            _obj.y *= _window.y
            _obj.ifft(usenifft=usenifft)
            x, y = find_roi(_obj.x, _obj.y)
            if not fastmath:
                self.Z_cont[:, idx] = y
            try:
                centx, _ = find_center(x, y)
                self.found_centers[_center] = centx
            except ValueError as err:
                self.errorcounter += 1
                if errors == "ignore":
                    self.found_centers[_center] = None
                else:
                    raise err
            if not silent:  # This creates about 5-15% overhead..
                sys.stdout.write('\r')
                j = (idx + 1) / winlen
                sys.stdout.write(
                    "Progress : [%-30s] %d%% (Errors: %d)" % ('=' * int(30 * j), 100 * j, self.errorcounter)
                )
                sys.stdout.flush()

    def _apply_window_seq_parallel(
            self, fastmath=True, usenifft=False, errors="ignore"
    ):
        self.errorcounter = 0
        if not fastmath:
            # here we setup the shape for the Z array and allocate Y, because
            # it is much faster than using np.append in every iteration
            _x, _y, _, _ = self._safe_cast()
            _obj = FFTMethod(_x, _y)
            _obj.ifft(usenifft=usenifft)
            x, y = find_roi(_obj.x, _obj.y)
            yshape = y.size
            self.Y_cont = np.array(x)
            xshape = len(self.window_seq)
            self.Z_cont = np.empty(shape=(yshape, xshape))

        for idx, (_center, _window) in enumerate(self.window_seq.items()):
            element = self._prepare_element(idx, _window, fastmath, usenifft, errors)
            if element is None:
                self.errorcounter += 1  # This might be useless, since we lazy evaluate things..
            self.found_centers[_center] = element

    @delayed
    def _prepare_element(self, idx, window, fastmath=True, usenifft=False, errors="ignore"):
        _x, _y, _, _ = self._safe_cast()
        _obj = FFTMethod(_x, _y)
        _obj.y *= window.y
        _obj.ifft(usenifft=usenifft)
        x, y = find_roi(_obj.x, _obj.y)
        if not fastmath:
            self.Z_cont[:, idx] = y
        try:
            centx, _ = find_center(x, y)
            return centx
        except ValueError as err:
            if errors == "ignore":
                return None
            else:
                raise err

    def _clean_centers(self, silent=False):
        dct = {k: v for k, v in self.found_centers.items() if v is not None}
        self.found_centers = dct
        winlen = len(self.window_seq)
        usefullen = len(self.found_centers)

        if not silent:
            if winlen != usefullen:
                print(
                    f"\nIn total {abs(winlen-usefullen)} out of {winlen} datapoints "
                    f"were thrown away due to ambiguous peak positions."
                )

    def errorplot(self, *args, **kwargs):
        """
        Plot the errors of fitting.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            An axis to draw the plot on. If not given, it will plot
            of the last used axis.
        percent : bool, optional
            Whether to plot percentage difference. Default is False.
        title : str, optional
            The title of the plot. Default is "Errors".
        kwargs : dict, optional
            Additional keyword arguments to pass to plot function.
        """
        try:
            getattr(self.GD, "errorplot", None)(*args, **kwargs)
        except TypeError:
            raise NotCalculatedException("Must calculate before plotting errors.")

    @property
    def get_GD(self):
        """
        Return the GD if it is already calculated.
        """
        if self.GD is not None:
            return self.GD
        raise NotCalculatedException("Must calculate GD first.")

    @property
    def errors(self):
        """
        Return the fitting errors as np.ndarray.
        """
        return getattr(self.GD, "errors", None)

    def _collect_failures(self):
        return [k for k in self.window_seq.keys() if k not in self.found_centers.keys()]

    def _construct_heatmap_data(self):
        self.X_cont = np.fromiter(self.window_seq.keys(), dtype=float)

    def heatmap(self, ax=None, levels=None, cmap="viridis", include_ridge=True):
        """
        Plot the heatmap.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            An axis to draw the plot on. If not given, it will plot
            of the last used axis.
        levels : np.ndarray, optional
            The levels to use for plotting.
        cmap : str, optional
            The colormap to use.
        include_ridge : bool, optional
            Whether to mark the detected ridge of the plot.
            Default is True.
        """
        if self.GD is None:
            raise NotCalculatedException("Must calculate GD first.")

        if self.fastmath:
            raise ValueError(
                "You need to recalculate with `fastmath=False` to plot the heatmap."
            )
        # Only construct if we need to..
        if not (self.Y_cont.size, self.X_cont.size) == self.Z_cont.shape:
            self._construct_heatmap_data()
        if ax is None:
            plt.contourf(
                self.X_cont, self.Y_cont, self.Z_cont, levels=levels, cmap=cmap, extend="both"
            )
        else:
            ax.contourf(
                self.X_cont, self.Y_cont, self.Z_cont, levels=levels, cmap=cmap, extend="both"
            )
        if include_ridge:
            if ax is None:
                plt.plot(*self.GD.data, color='red', label='detected ridge')
            else:
                ax.plot(*self.GD.data, color='red', label='detected ridge')
            plt.legend()
        if ax is None:
            plt.xlabel('Window center [PHz]')
            plt.ylabel('Delay [fs]')
            try:
                upper_bound = min(1.5 * np.max(self.GD.data[1]), np.max(self.Y_cont))
                plt.ylim(None, upper_bound)
            except ValueError:
                pass
            # ValueError: zero-size array to reduction operation maximum which has no identity
            # This means that the array is empty, we should pass that case.
        else:
            ax.set_autoscalex_on(False)
            try:
                upper_bound = min(1.5 * np.max(self.GD.data[1]), np.max(self.Y_cont))
                ax.set(
                    xlabel="Window center [PHz]",
                    ylabel="Delay [fs]",
                    ylim=(None, upper_bound)
                )
            except ValueError:
                pass

    def get_heatmap_data(self):
        """
        Return the data which was used to create the heatmap.

        Returns
        -------
        X_cont : np.ndarray
            The window centers with shape (n,).
        Y_cont : np.ndarray
            The time axis calculated from the IFFT of the dataset with shape (m,).
        Z_cont : np.ndarray
            2D array with shape (m, n) containing the depth information.
        """
        if all([self.Y_cont.size != 0, self.Z_cont.size != 0]):
            self._construct_heatmap_data()
        else:
            raise ValueError("Must calculate with `fastmath=False` before trying to access the heatmap data.")

        return self.X_cont, self.Y_cont, self.Z_cont



import unittest
from unittest import TestCase
from pysprint.core.nufft import nuifft

import numpy as np


class TestNuifft(TestCase):
    def setUp(self):
        rng = np.random.RandomState(151560)
        self.x = 100 * rng.rand(100)
        self.y = np.exp(1j * self.x)
        self.gl = len(self.x)
        self.eps = 1e-12

    def test_nuifft(self):
        dft = nuifft(self.x, self.y, self.gl, exponent="positive")
        fft = nuifft(self.x, self.y, self.gl, exponent="positive", epsilon=self.eps)
        np.testing.assert_allclose(dft, fft, rtol=self.eps ** 0.95)

        dft = nuifft(self.x, self.y.real, self.gl, exponent="positive")
        fft = nuifft(
            self.x, self.y.real, self.gl, exponent="positive", epsilon=self.eps
        )
        np.testing.assert_allclose(dft, fft, rtol=self.eps ** 0.95)


if __name__ == "__main__":
    unittest.main()



import re
from pathlib import Path
from typing import IO, AnyStr, List, Optional, Union
import warnings

import pandas as pd

from pysprint.utils.exceptions import PySprintWarning

PathOrBuffer = Union[str, Path, IO[AnyStr]]


def _parse_single_line(line: str):
    possible_separators = (":", "_", "-", "\t", " ", ";", ",", ".")
    idx = 0
    parsed_line = (line,)
    while idx < len(possible_separators):
        res = line.split(possible_separators[idx])
        if len(res) == 2:
            parsed_line = res
            break
        idx += 1

    # if we fail with the possible separators, let's see if
    # we can split up on contiguous whitespace to two pieces.

    if len(parsed_line) == 1:
        ln = re.split(r'\s+', *parsed_line)
        if len(ln) == 2:
            parsed_line = ln
    rx = '[' + re.escape(''.join(possible_separators[2:-4])) + ']'

    # remove the obvious unwanted chars
    parsed_line = [element.strip("\n").strip("\x00") for element in parsed_line]

    # remove the separator-like chars
    parsed_line = [re.sub(rx, "", el) for el in parsed_line]

    # remove whitespace around elements
    parsed_line = [element.strip(" ") for element in parsed_line]

    return parsed_line


def _parse_metadata(filename, ref=None, sam=None, meta_len=1, encoding='utf-8'):
    _meta = {}
    with open(filename, encoding=encoding) as file:
        comm = next(file).strip("\n").split("-")[-1].lstrip(" ")
        additional = [
            _parse_single_line(next(file))
            for _ in range(1, meta_len + 1)
        ]
        if meta_len != 0:
            _meta = {"comment": comm}
        for el in additional:
            if len(el) == 2:
                _meta[el[0]] = el[1]
            else:
                # _meta.setdefault("unparsed", []).append(el)
                # should do the trick here
                if "unparsed" not in _meta:
                    _meta["unparsed"] = []
                _meta["unparsed"].append(el)
    if ref is not None:
        with open(ref, encoding=encoding) as file:
            comm = next(file).strip("\n").split("-")[-1].lstrip(" ")
            if meta_len != 0:
                _meta["reference_comment"] = comm
    if sam is not None:
        with open(sam, encoding=encoding) as file:
            comm = next(file).strip("\n").split("-")[-1].lstrip(" ")
            if meta_len != 0:
                _meta["sample_comment"] = comm
    return _meta


def _parse_raw(
        filename: PathOrBuffer,
        ref: Optional[PathOrBuffer] = None,
        sam: Optional[PathOrBuffer] = None,
        skiprows: int = 0,
        decimal: str = ".",
        sep: Optional[str] = None,
        delimiter: Optional[str] = None,
        comment: Optional[str] = None,
        usecols: Optional[List] = None,
        names=None,
        swapaxes=False,
        na_values=None,
        skip_blank_lines=True,
        keep_default_na=False,
        meta_len: int = 1
):
    if len([x for x in (ref, sam) if x is not None]) == 1:
        warnings.warn(
            "Reference and sample arm should be passed together or neither one.",
            PySprintWarning
        )
    if sep and delimiter:
        warnings.warn("Specified both `sep` and `delimiter`, using `delimiter` as default.")
        separator = delimiter
    else:
        separator = sep or delimiter or ","

    if comment is None:
        comment = "#"

    if usecols is None:
        usecols = [0, 1]

    if names is None:
        names = ["x", "y"]

    if skiprows < meta_len:
        warnings.warn(
            f"Skiprows is currently {skiprows}, but"
            f" meta information is set to {meta_len}"
            " lines. This implies that either one is probably wrong.",
            PySprintWarning,
        )

    _meta = _parse_metadata(filename, ref, sam, meta_len=meta_len)

    df = pd.read_csv(
        filename,
        skiprows=skiprows,
        sep=separator,
        decimal=decimal,
        usecols=usecols,
        names=names,
        comment=comment,
        na_values=na_values,
        skip_blank_lines=skip_blank_lines,
        keep_default_na=keep_default_na
    )
    if ref is not None and sam is not None:
        r = pd.read_csv(
            ref,
            skiprows=skiprows,
            sep=separator,
            decimal=decimal,
            usecols=usecols,
            names=names,
            comment=comment,
            na_values=na_values,
            skip_blank_lines=skip_blank_lines,
            keep_default_na=keep_default_na
        )
        s = pd.read_csv(
            sam,
            skiprows=skiprows,
            sep=separator,
            decimal=decimal,
            usecols=usecols,
            names=names,
            comment=comment,
            na_values=na_values,
            skip_blank_lines=skip_blank_lines,
            keep_default_na=keep_default_na
        )

        if swapaxes:
            return {
                "x": df["y"].values,
                "y": df["x"].values,
                "ref": r["x"].values,
                "sam": s["x"].values,
                "meta": _meta
            }
        return {
            "x": df["x"].values,
            "y": df["y"].values,
            "ref": r["y"].values,
            "sam": s["y"].values,
            "meta": _meta
        }
    if swapaxes:
        return {
            "x": df["y"].values,
            "y": df["x"].values,
            "meta": _meta
        }
    return {
        "x": df["x"].values,
        "y": df["y"].values,
        "meta": _meta
    }
